use crate::entity::{Entity, EntityDatum, EntityVisitor};

pub struct Unique<T: EntityDatum> {
    _ghost: std::marker::PhantomData<T>,
}

impl<T: EntityDatum> EntityDatum for Unique<T> {}

pub struct IDMap<T: Entity> {
    conn: std::cell::OnceCell<std::sync::Arc<sqlite::ConnectionThreadSafe>>,
    _ghost: std::marker::PhantomData<T>,
}

pub struct AssocSet<T: Entity> {
    conn: std::cell::OnceCell<std::sync::Arc<sqlite::ConnectionThreadSafe>>,
    _ghost: std::marker::PhantomData<T>,
}

pub struct Index<T: Entity, Key: EntityDatum> {
    conn: std::cell::OnceCell<std::sync::Arc<sqlite::ConnectionThreadSafe>>,
    _ghost: std::marker::PhantomData<(T, Key)>,
}

pub trait DatabaseItem {
    fn item_key() -> &'static str;
    fn dependency_keys() -> &'static [&'static str];

    fn is_index() -> bool {
        false
    }
    fn index_over() -> &'static str {
        unreachable!()
    }
    fn index_columns() -> &'static [&'static str] {
        unreachable!()
    }

    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

pub trait DatabaseItemVisitor {
    fn visit<DI: DatabaseItem>(&mut self)
    where
        Self: Sized;
}

pub trait DatabaseSpec {
    fn give_connection(&mut self, conn: std::sync::Arc<sqlite::ConnectionThreadSafe>);
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

impl<T: Entity> DatabaseSpec for IDMap<T> {
    fn give_connection(&mut self, conn: std::sync::Arc<sqlite::ConnectionThreadSafe>) {
        self.conn
            .set(conn)
            .ok()
            .expect("couldn't set once_cell with sqlite connection!");
    }
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor) {
        visitor.visit::<T>()
    }
}

pub trait Database {
    fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
    where
        Self: Sized;
}

#[cfg(test)]
mod simple_test {
    use super::*;
    use crate::entity::{EntityPart, EntityPartVisitor};
    // simple hand-built database example

    struct SimpleEntity {
        name: Unique<String>,
    }

    // normally this would be invisible, but for testing purposes we expose it
    impl SimpleEntity {
        // pub const Name: SimpleEntityName = SimpleEntityName;
    }
    struct SimpleEntityName;
    impl EntityPart for SimpleEntityName {
        type Datum = String;
        fn part_name() -> &'static str {
            "name"
        }
        fn is_unique() -> bool {
            true
        }
    }

    impl Entity for SimpleEntity {
        fn entity_name() -> &'static str {
            "simple_entity"
        }
        fn accept_part_visitor(visitor: &mut impl EntityPartVisitor) {
            visitor.visit::<SimpleEntityName>();
        }
    }

    struct SimpleDatabase {
        strings: IDMap<SimpleEntity>,
    }

    impl Database for SimpleDatabase {
        fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
        where
            Self: Sized,
        {
            struct SimpleDatabaseStringsItem;
            impl DatabaseItem for SimpleDatabaseStringsItem {
                fn item_key() -> &'static str {
                    "strings"
                }
                fn dependency_keys() -> &'static [&'static str] {
                    &[]
                }

                fn accept_entity_visitor(visitor: &mut impl EntityVisitor) {
                    visitor.visit::<SimpleEntity>();
                }
            }

            visitor.visit::<SimpleDatabaseStringsItem>();
        }
    }

    #[test]
    fn part_visitor() {
        struct V {
            v: Vec<std::any::TypeId>,
        }
        impl EntityPartVisitor for V {
            fn visit<EP: EntityPart>(&mut self) {
                self.v.push(std::any::TypeId::of::<EP>());
            }
        }

        let mut vis = V { v: vec![] };
        SimpleEntity::accept_part_visitor(&mut vis);
        assert_eq!(
            vis.v.as_slice(),
            &[std::any::TypeId::of::<SimpleEntityName>()]
        );
    }
}

#[cfg(test)]
mod derive_test {
    use microrm_macros::{Database, Entity};

    use super::IDMap;
    #[derive(Entity)]
    struct SimpleEntity {
        name: super::Unique<String>,
    }

    #[derive(Database)]
    struct SimpleDB {
        entities: IDMap<SimpleEntity>,
    }

    #[test]
    fn collect_test() {
        crate::schema::collect_from_database::<SimpleDB>();
    }
}

/*
// pub trait

pub struct Skill { }
impl Entity for Skill {}

pub struct MenuItem {
    name: String,
    required_skills: Set<Skill>,
    cost: f32,
}

impl Entity for MenuItem {}

struct EmployeeShiftMap;

pub struct Employee {
    assigned_shifts: NamedMap<EmployeeShiftMap, Shift>,
}

impl Entity for Employee {}

pub struct Shift {
    name: String,
    employees: Set<Employee>,
}

impl Entity for Shift {}

pub struct FoodTruckSchema {
    menu: Collection<MenuItem>,
}
*/
